import struct
import time as T
import datetime
import base58
import binascii
import hashlib
import urllib.request
import json

def get_record(url):
    #获取对应url中的json数据
    resp = urllib.request.urlopen(url)
    ele_json = json.loads(resp.read())
    return ele_json

def get_address(hash, no):
    #通过API获取具体transaction中的address和value
    url = 'https://api.blockcypher.com/v1/btc/main/txs/%s'%hash
    print(url)
    ele_json = get_record(url)
    address = ele_json['outputs'][no]['addresses'][0]
    value = ele_json['outputs'][no]['value']
    return address,value

def loadFile(path):
    #将path中的block存入filepath的json文件中
    f = open(path,'rb+')
    BlockNum = 1
    #filePath = ""
    #while(f!=""):
    while(BlockNum<=10):
        filePath = "E:/Bitcoin/Blocks/json/block%d.json"%BlockNum
        print("BlockNum: %d"%BlockNum)
        NewBlock = Block(f)
        if (BlockNum >= 5):
            NewBlock.Savejson(filePath)
        BlockNum+=1
    f.close()

def uint1(stream):
    return ord(stream.read(1))
def uint2(stream):
    return struct.unpack('H',stream.read(2))[0]
def uint4(stream):
    return struct.unpack('I',stream.read(4))[0]
def uint8(stream):
    return struct.unpack('Q',stream.read(8))[0]
def hash32(stream):
    return stream.read(32)[::-1]
def time(stream):
    time = uint4(stream)
    return time
def varint(stream):
    size = uint1(stream)
    if size < 0xfd:
        return size
    elif size == 0xfd:
        return uint2(stream)
    elif size == 0xfe:
        return uint4(stream)
    elif size == 0xff:
        return uint8(stream)
    return -1

def hashStr(bytebuffer):
    Str = ''
    for i in bytebuffer:
        temp = str(hex(i)[2:])
        if len(temp) == 1:
            Str += "0"
        Str += temp
    return Str

class Block:
    def __init__(self, blockchain):
        self.magicNum = uint4(blockchain)
        self.blocksize = uint4(blockchain)
        self.setHeader(blockchain)
        self.txCount = varint(blockchain)
        self.Txs = []
        for i in range(0, self.txCount):
            tx = Tx(blockchain)
            self.Txs.append(tx)
    def setHeader(self,blockchain):
        self.blockHeader = BlockHeader(blockchain)

    def Savejson(self,filePath):
        #将block中的数据保存入filePath路径的json文件中，dat文件读取与API相结合，
        #先读取本block中的时间，在读取block中的input的prehash，通过API读取prehash对应的output中的address和value，
        #最后读取本blockoutput中的address和value。存入json文件
        jsonFile = open(filePath,'w')

        Block_json = {
            "time":self.blockHeader.time,
            "Txs":[]
        }
        for t in self.Txs:
            flag = True
            tx = {
                "inputs":[],
                "outputs":[]
            }
            for input in t.inputs:
                prevhash = hashStr(input.prevhash)
                no = input.txOutId
                if prevhash == "0000000000000000000000000000000000000000000000000000000000000000":
                    address = None
                    value = 0
                else:
                    try:
                        address,value = get_address(prevhash,no)
                    except:
                        print("Error:",prevhash,'\t',no)
                        flag = False
                        break
                i = {
                    "address":address,
                    "value":value
                }
                tx["inputs"].append(i)
                T.sleep(10)

            if (flag == False):
                continue

            for output in t.outputs:
                try:
                    address,value = output.get_address()
                except:
                    flag = False
                    break
                o = {
                    "address":address,
                    "value":value
                }
                tx["outputs"].append(o)

            if (flag == False):
                continue

            Block_json["Txs"].append(tx)

        print(Block_json)
        json.dump(Block_json,jsonFile)
        jsonFile.close()

class BlockHeader:
    #配合block类进行初始化
    def __init__(self,blockchain):
        self.version = uint4(blockchain)
        self.previousHash = hash32(blockchain)
        self.merkleHash = hash32(blockchain)
        self.time = uint4(blockchain)

        timeArray = T.localtime(self.time)#1970秒数
        otherStyleTime = T.strftime("%Y-%m-%d %H:%M:%S", timeArray)
        print(otherStyleTime)

        self.bits = uint4(blockchain)
        self.nonce = uint4(blockchain)

class Tx:
    #配合block类进行初始化。用到了txInput类和txOutput类
    def __init__(self,blockchain):
        self.version = uint4(blockchain)
        self.inCount = varint(blockchain)
        self.inputs = []
        for i in range(0,self.inCount):
            input = txInput(blockchain)
            self.inputs.append(input)
        self.outCount = varint(blockchain)
        self.outputs = []
        if self.outCount>0:
            for i in range(0,self.outCount):
                output = txOutput(blockchain)
                self.outputs.append(output)
        self.lockTime = uint4(blockchain)

class txInput:
    #配合Tx类
    def __init__(self,blockchain):
        self.prevhash = hash32(blockchain)
        self.txOutId = uint4(blockchain)
        self.scriptLen = varint(blockchain)
        self.scriptSig = blockchain.read(self.scriptLen)
        self.seqNo = uint4(blockchain)

class txOutput:
    #配合Tx类
    def __init__(self,blockchain):
        self.value = uint8(blockchain)
        self.scriptLen = varint(blockchain)
        self.pubkey = blockchain.read(self.scriptLen)

    def get_address(self):
        if (self.scriptLen == 67):
            address = in_address(hashStr(self.pubkey)[2:-2])
            print(hashStr(self.pubkey)[2:-2])
        elif (self.scriptLen == 25):
            address = compress_address(hashStr(self.pubkey)[6:-4])
            print(hashStr(self.pubkey)[6:-4])
        else:
            address = compress_address(hashStr(self.pubkey)[6:-4])
            print("Something Wrong!")
        return address,self.value

def in_address(in_pubkey):
    #将public key转化为地址
    #本函数绝对正确，但上面的获取pubkey过程不知道对不对
    in_pubkey_hash256 = hashlib.sha256(binascii.unhexlify(in_pubkey.encode())).hexdigest()
    in_pubkey_hash160 = hashlib.new('ripemd160',binascii.unhexlify(in_pubkey_hash256.encode())).hexdigest()
    in_verandpubkey_hash160 = "00" + in_pubkey_hash160
    in_verandpubkey_hash160_hash256 = hashlib.sha256(binascii.unhexlify(in_verandpubkey_hash160.encode())).hexdigest()
    in_verandpubkey_hash160_hash256_2 = hashlib.sha256(binascii.unhexlify(in_verandpubkey_hash160_hash256.encode())).hexdigest()
    in_subaddress = "00" + in_pubkey_hash160 + in_verandpubkey_hash160_hash256_2[:8]
    address  = base58.b58encode(binascii.unhexlify(in_subaddress.encode())).decode()
    return address

def compress_address(pubkey):
    #对长40位的16进制已处理pubkey使用，将其转化为地址
    in_verandpubkey_hash160 = "00" + pubkey
    in_verandpubkey_hash160_hash256 = hashlib.sha256(binascii.unhexlify(in_verandpubkey_hash160.encode())).hexdigest()
    in_verandpubkey_hash160_hash256_2 = hashlib.sha256(binascii.unhexlify(in_verandpubkey_hash160_hash256.encode())).hexdigest()
    in_subaddress = "00" + pubkey + in_verandpubkey_hash160_hash256_2[:8]
    address  = base58.b58encode(binascii.unhexlify(in_subaddress.encode())).decode()
    return address

path = "E:/Bitcoin/Blocks/blocks/blk00003.dat"
#filePath = "E:/Bitcoin/Blocks/files/blockchainTry4.txt"
loadFile(path)
